import logging
from abc import ABC, abstractmethod
from pathlib import Path

from lxml import etree

logger = logging.getLogger(__name__)


class XmlElement(ABC):
    optional = ()

    def __init__(self, element):
        self.path = None
        if isinstance(element, str | Path):
            assert Path(element).exists(), f"Invalid path {element}"
            self.path = element
            element = etree.parse(str(element)).getroot()
        assert isinstance(element, etree._Element), "Not an XML element"
        self.element = element
        self._data = None

    @property
    def data(self):
        if self._data is None:
            self._data = self.parse()
        return self._data

    def __getattr__(self, name):
        if self._data is None or name not in self._data:
            raise AttributeError
        return self._data[name]

    def _find(self, path, many=False):
        result = self.element.xpath(path, namespaces=self.namespaces)
        if isinstance(result, list) and not many:
            if not result:
                return
            return result[0]
        return result

    def get_text(self, path, dump=False):
        child = self._find(path)
        if child is None:
            return None
        elif isinstance(child, str):
            return child

        # Dump full content with children
        if dump:
            return etree.tostring(child, pretty_print=True).decode("utf-8")

        # Simply get inner text
        return child.text

    def get_float(self, path):
        text = self.get_text(path)
        if not text:
            return None
        return float(text)

    def get_int(self, path):
        text = self.get_text(path)
        if not text:
            return None
        return int(text)

    def get_instance(self, cls, path, many=False):
        child = self._find(path, many)
        if child is None:
            return None

        if many:
            return list(cls(item) for item in child)

        return cls(child)

    @abstractmethod
    def parse(self):
        """
        Returns a dict holding parsed data from XML.
        """

    def as_dict(self):
        # Circular dependencies
        from transkribus.pagexml import Tag

        def _dump(value):
            if isinstance(value, XmlElement | Tag):
                return value.as_dict()
            if isinstance(value, list | tuple):
                return list(map(_dump, value))
            return value

        return {name: _dump(value) for name, value in self.data.items()}



from pathlib import Path

import pytest

from teklia_toolbox.pagexml import PageXmlPage

SAMPLES = Path(__file__).parent / "samples"


def test_simple():
    page_xml = PageXmlPage(SAMPLES / "simple.xml")

    assert page_xml.metadata.creator == "TRP"
    assert page_xml.metadata.created == "2017-07-19T13:58:10.738+02:00"
    assert page_xml.metadata.last_change == "2017-07-19T14:04:22.502+02:00"

    assert page_xml.page.image_name == "FRAN_0021_0023_L.jpg"
    assert page_xml.page.image_width == "3195"
    assert page_xml.page.image_height == "3731"
    assert page_xml.page.ordering == ["TextRegion_1500465748446_12"]
    assert len(page_xml.page.text_regions) == 1

    region = page_xml.page.text_regions[0]
    assert region.id == "TextRegion_1500465748446_12"
    assert region.points == [(2974, 1270), (16, 1270), (18, 105), (2976, 105)]
    assert len(region.tags) == 1
    assert region.tags[0].name == "readingOrder"
    assert region.tags[0].index == "0"


def test_2_regions():
    page_xml = PageXmlPage(SAMPLES / "2_regions.xml")

    assert len(page_xml.page.text_regions) == 2
    assert page_xml.page.ordering == [
        "TextRegion_1522155769847_482",
        "TextRegion_1522155769847_481",
    ]
    region_1, region_2 = page_xml.page.text_regions

    assert region_1.id == "TextRegion_1522155769847_482"
    assert region_2.id == "TextRegion_1522155769847_481"
    assert len(region_1.tags) == 1
    assert region_1.tags[0].name == "readingOrder"
    assert region_1.tags[0].index == "0"
    assert len(region_2.tags) == 1
    assert region_2.tags[0].name == "readingOrder"
    assert region_2.tags[0].index == "1"
    assert region_1.points == [(3509, 1675), (0, 1675), (0, 0), (3509, 0)]
    assert region_2.points == [(3509, 4569), (0, 4569), (0, 1675), (3509, 1675)]

    assert page_xml.page.sort_regions(page_xml.page.text_regions) == [
        region_1,
        region_2,
    ]


def test_no_reading_order():
    page_xml = PageXmlPage(SAMPLES / "no_reading_order.xml")

    assert len(page_xml.page.text_regions) == 3
    assert page_xml.page.ordering == []
    region_1, region_2, region_3 = page_xml.page.text_regions

    assert region_1.id == "TextRegion_1522155769847_482"
    assert region_2.id == "TextRegion_1522155769847_481"
    assert region_3.id == "TextRegion_1522155769847_484"

    assert (
        page_xml.page.sort_regions(page_xml.page.text_regions)
        == page_xml.page.text_regions
    )


def test_unordered():
    page_xml = PageXmlPage(SAMPLES / "unordered.xml")

    assert len(page_xml.page.text_regions) == 3
    assert page_xml.page.ordering == [
        "TextRegion_1522155769847_482",
        "TextRegion_1522155769847_481",
        "TextRegion_1522155769847_484",
    ]

    region_1, region_2, region_3 = page_xml.page.text_regions
    assert region_1.id == "TextRegion_1522155769847_482"
    assert region_2.id == "TextRegion_1522155769847_481"
    assert region_3.id == "TextRegion_1522155769847_484"

    assert page_xml.page.sort_regions(page_xml.page.text_regions) == [
        region_1,
        region_2,
        region_3,
    ]


def test_transcript():
    page_xml = PageXmlPage(SAMPLES / "transcript.xml")

    assert len(page_xml.page.relations) == 1
    assert page_xml.page.relations[0] == [
        "TextRegion_1540299380975_9",
        "TextRegion_1540299473514_23",
    ]
    assert len(page_xml.page.text_regions) == 2
    region_1, region_2 = page_xml.page.text_regions

    assert region_1.id == "TextRegion_1540299380975_9"
    assert region_1.points == [(12, 34), (56, 78), (910, 1112)]
    assert region_1.text == "B .1\nLouis Joseph\nPierre Siméon\nLemieux"
    assert len(region_1.tags) == 2
    assert region_1.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "0",
    }
    assert region_1.tags[1].as_dict() == {
        "name": "structure",
        "type": "marginalia",
    }

    assert len(region_1.lines) == 4
    line_1, line_2, line_3, line_4 = region_1.lines

    assert line_1.id == "r1l6"
    assert len(line_1.tags) == 2
    assert line_1.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "0",
    }
    assert line_1.tags[1].as_dict() == {
        "name": "structure",
        "type": "ref",
    }
    assert line_1.points == [(12, 34), (56, 78), (910, 1112)]
    assert line_1.baseline == [(13, 37), (42, 42), (37, 13)]
    assert line_1.text == "B .1"

    assert line_2.id == "r1l7"
    assert len(line_2.tags) == 2
    assert line_2.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "1",
    }
    assert line_2.tags[1].as_dict() == {
        "name": "_prenom",
        "offset": "0",
        "length": "12",
        "continued": "true",
        "_role": "sujet",
    }
    assert line_2.points == [(12, 34), (56, 78), (910, 1112)]
    assert line_2.baseline == [(13, 37), (42, 42), (37, 13)]
    assert line_2.text == "Louis Joseph"

    assert line_3.id == "r1l8"
    assert len(line_3.tags) == 2
    assert line_3.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "2",
    }
    assert line_3.tags[1].as_dict() == {
        "name": "_prenom",
        "offset": "0",
        "length": "13",
        "continued": "true",
        "_role": "sujet",
    }
    assert line_3.points == [(12, 34), (56, 78), (910, 1112)]
    assert line_3.baseline == [(13, 37), (42, 42), (37, 13)]
    assert line_3.text == "Pierre Siméon"

    assert line_4.id == "r1l9"
    assert len(line_4.tags) == 2
    assert line_4.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "3",
    }
    assert line_4.tags[1].as_dict() == {
        "name": "_nom",
        "offset": "0",
        "length": "7",
    }
    assert line_4.points == [(12, 34), (56, 78), (910, 1112)]
    assert line_4.baseline == [(13, 37), (42, 42), (37, 13)]
    assert line_4.text == "Lemieux"

    assert region_2.id == "TextRegion_1540299473514_23"
    assert len(region_2.tags) == 1
    assert region_2.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "1",
    }
    assert region_2.points == [(12, 34), (56, 78), (910, 1112)]
    assert (
        region_2.text
        == "Le onze janvier mil neuf centsept\nnous prêtre soussigné avons baptisé Louis"
    )
    assert len(region_2.lines) == 2
    line_1, line_2 = region_2.lines

    assert line_1.id == "r2l12"
    assert len(line_1.tags) == 2
    assert line_1.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "0",
    }
    assert line_1.tags[1].as_dict() == {
        "name": "_date",
        "offset": "3",
        "length": "30",
        "_enregistrement": "1",
    }
    assert line_1.points == [(12, 34), (56, 78), (910, 1112)]
    assert line_1.baseline == [(13, 37), (42, 42), (37, 13)]
    assert line_1.confidence is None
    assert line_1.text == "Le onze janvier mil neuf centsept"

    assert line_2.id == "r2l13"
    assert len(line_2.tags) == 2
    assert line_2.tags[0].as_dict() == {
        "name": "readingOrder",
        "index": "1",
    }
    assert line_2.tags[1].as_dict() == {
        "name": "_prenom",
        "offset": "36",
        "length": "5",
        "continued": "true",
        "_role": "sujet",
    }
    assert line_2.points == [(12, 34), (56, 78), (910, 1112)]
    assert line_2.baseline == [(13, 37), (42, 42), (37, 13)]
    assert line_2.confidence == 0.42
    assert line_2.text == "nous prêtre soussigné avons baptisé Louis"


def test_no_namespace():
    with pytest.raises(
        Exception, match="There are no URLs, the PAGE namespace is missing"
    ):
        PageXmlPage(SAMPLES / "no_namespace.xml")


def test_multiple_versions():
    with pytest.raises(
        Exception, match="can't manage different versions at the same time"
    ):
        PageXmlPage(SAMPLES / "multiple_versions.xml")


def test_different_version():
    page_xml = PageXmlPage(SAMPLES / "different_version.xml")
    assert page_xml.page.image_name == "0_f546b_default.jpg"
    assert page_xml.page.image_width == "3509"
    assert page_xml.page.image_height == "4905"
    assert len(page_xml.page.text_regions) == 1

    region = page_xml.page.text_regions[0]
    assert len(region.lines) == 2
    line_1, line_2 = region.lines
    assert line_1.points == [
        (830, 1479),
        (834, 1389),
        (907, 1373),
        (948, 1389),
        (1006, 1357),
    ]
    assert line_1.baseline == [(830, 1479), (2592, 1467)]
    assert line_1.text == "LA BIBLIOTHÈQUE NATIONALE EN 1984"

    assert line_2.points == [
        (862, 2333),
        (879, 2178),
        (985, 2207),
        (1042, 2190),
        (1100, 2219),
    ]
    assert line_2.baseline == [(862, 2333), (2539, 2333)]
    assert line_2.text == "RAPPORT D'ACTIVITÉ"



import tempfile
from pathlib import Path
from unittest import TestCase
from unittest.mock import patch

from teklia_toolbox.config import ConfigParser, ConfigurationError, dir_path, file_path


class TestConfigParser(TestCase):
    @patch("teklia_toolbox.config._all_checks")
    def test_file_path(self, all_checks_mock):
        all_checks_mock.return_value = True

        with self.assertRaisesRegex(AssertionError, " does not exist"):
            file_path("/aaaaaaa")

        with tempfile.NamedTemporaryFile() as f:
            parent_path = Path(f.name).parent
            with self.assertRaisesRegex(AssertionError, " is not a file"):
                file_path(parent_path)

            self.assertEqual(file_path(f.name), Path(f.name))

            # Existence checks should be ignored without all_checks
            all_checks_mock.return_value = False
            self.assertEqual(file_path(parent_path), parent_path)
            self.assertEqual(file_path("/aaaaaaa"), Path("/aaaaaaa"))

    @patch("teklia_toolbox.config._all_checks")
    def test_dir_path(self, all_checks_mock):
        all_checks_mock.return_value = True
        with tempfile.TemporaryDirectory() as d:
            self.assertEqual(dir_path(d), Path(d))

        with self.assertRaisesRegex(AssertionError, " does not exist"):
            dir_path("/aaaaaaa")

        with tempfile.NamedTemporaryFile() as f:
            with self.assertRaisesRegex(AssertionError, " is not a directory"):
                dir_path(f.name)

            # Existence checks should be ignored without all_checks
            all_checks_mock.return_value = False
            self.assertEqual(dir_path(f.name), Path(f.name))
            self.assertEqual(dir_path("/aaaaaaa"), Path("/aaaaaaa"))

    def test_configuration_error(self):
        error = ConfigurationError({"a": "b"})
        self.assertDictEqual(error.errors, {"a": "b"})
        self.assertEqual(str(error), '{"a": "b"}')
        self.assertEqual(repr(error), 'ConfigurationError({"a": "b"})')

    def test_add_option(self):
        parser = ConfigParser()
        parser.add_option("test", type=int)
        with self.assertRaisesRegex(
            AssertionError, "test is an already defined option"
        ):
            parser.add_option("test")
        with self.assertRaisesRegex(AssertionError, "Option type must be callable"):
            parser.add_option("toast", type=...)

    def test_parse_not_found(self):
        parser = ConfigParser()
        parser.add_option("something", default="thing")
        with self.assertRaises(FileNotFoundError):
            parser.parse(Path("/aaaaaaa"))
        self.assertDictEqual(
            parser.parse(Path("/aaaaaaa"), exist_ok=True),
            {"something": "thing"},
        )

    def test_parse_no_extra(self):
        parser = ConfigParser(allow_extra_keys=False)
        parser.add_option("something", type=bool, default=True)
        with self.assertRaises(ConfigurationError) as context:
            parser.parse_data({"something": False, "unknownthing": "oops"})

        self.assertDictEqual(
            context.exception.errors,
            {
                "unknownthing": "This option does not exist",
            },
        )

    def test_parse_null_no_default(self):
        parser = ConfigParser()
        parser.add_option("something", type=str)
        with self.assertRaises(ConfigurationError) as context:
            parser.parse_data({"something": None})

        self.assertDictEqual(
            context.exception.errors,
            {
                "something": "This option is required",
            },
        )

    def test_parse_null_with_default(self):
        parser = ConfigParser()
        parser.add_option("something", type=str, default=None)
        self.assertDictEqual(
            parser.parse_data({"something": None}),
            {"something": None},
        )



import datetime
from timeit import default_timer


class Timer:
    """
    A context manager to help measure execution times
    """

    def __init__(self):
        self.timer = default_timer

    def __enter__(self):
        self.start = self.timer()
        return self

    def __exit__(self, *args):
        end = self.timer()
        self.elapsed = end - self.start
        self.delta = datetime.timedelta(seconds=self.elapsed)



import json
import os
import sys
from collections import namedtuple
from collections.abc import Mapping
from pathlib import Path

import yaml

Option = namedtuple("Option", ["type", "default"])


# Used as a default value in `ConfigParser.add_option(default=UNSET)`
# because default=None implies that the option is optional
UNSET = object()


def _all_checks():
    """
    Prevents checking for path existence when running unit tests
    or other dev-related operations.
    This is the same as settings.ALL_CHECKS, but since the configuration
    is accessed before settings are initialized, it has to be copied here.
    This is made as a method to make mocking in unit tests much simpler
    than with a module-level constant.
    """
    os.environ.get("ALL_CHECKS") == "true" or "runserver" in sys.argv


def file_path(data):
    path = Path(data).resolve()
    if _all_checks():
        assert path.exists(), f"{path} does not exist"
        assert path.is_file(), f"{path} is not a file"
    return path


def dir_path(data):
    path = Path(data).resolve()
    if _all_checks():
        assert path.exists(), f"{path} does not exist"
        assert path.is_dir(), f"{path} is not a directory"
    return path


class ConfigurationError(ValueError):
    def __init__(self, errors, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.errors = errors

    def __str__(self):
        return json.dumps(self.errors)

    def __repr__(self):
        return f"{self.__class__.__name__}({self!s})"


class ConfigParser:
    def __init__(self, allow_extra_keys=True):
        """
        :param allow_extra_keys bool: Ignore extra unspecified keys instead
        of causing errors.
        """
        self.options = {}
        self.allow_extra_keys = allow_extra_keys

    def add_option(self, name, *, type=str, many=False, default=UNSET):
        assert name not in self.options, f"{name} is an already defined option"
        assert callable(type), "Option type must be callable"
        if many:
            self.options[name] = Option(lambda data: list(map(type, data)), default)
        else:
            self.options[name] = Option(type, default)

    def add_subparser(self, *args, allow_extra_keys=True, **kwargs):
        """
        Add a parser as a new option to this parser,
        to allow finer control over nested configuration options.
        """
        parser = ConfigParser(allow_extra_keys=allow_extra_keys)
        self.add_option(*args, **kwargs, type=parser.parse_data)
        return parser

    def parse_data(self, data):
        """
        Parse configuration data from a dict.
        Will raise ConfigurationError if any error is detected.
        """
        if not isinstance(data, Mapping):
            raise ConfigurationError("Parser data must be a mapping")

        parsed, errors = {}, {}

        if not self.allow_extra_keys:
            for name in data:
                if name not in self.options:
                    errors[name] = "This option does not exist"
        for name, option in self.options.items():
            if data.get(name) is not None:
                value = data[name]
            elif option.default is UNSET:
                errors[name] = "This option is required"
                continue
            elif option.default is None:
                parsed[name] = None
                continue
            else:
                value = option.default

            try:
                parsed[name] = option.type(value)
            except ConfigurationError as e:
                # Allow nested error dicts for nicer error messages
                # with add_subparser
                errors[name] = e.errors
            except Exception as e:
                errors[name] = str(e)

        if errors:
            raise ConfigurationError(errors)
        return parsed

    def parse(self, path, exist_ok=False):
        if not path.is_file() and exist_ok:
            # Act like the file is empty
            return self.parse_data({})
        return self.parse_data(yaml.safe_load(path.read_bytes()))



import logging
from pathlib import Path
from urllib.parse import urlparse

import requests
import urllib3
from tenacity import (
    retry,
    retry_if_exception_type,
    stop_after_attempt,
    wait_fixed,
)

from arkindex import ArkindexClient

logger = logging.getLogger(__name__)


def should_verify_cert(url: str) -> bool:
    """
    Skip SSL certification validation when hitting a development instance
    """
    if not url:
        return True

    host = urlparse(url).netloc
    return not host.endswith("ark.localhost")


def get_arkindex_client(**options) -> ArkindexClient:
    # Skip SSL verification in Arkindex API client for local development hosts
    verify = should_verify_cert(options.get("base_url"))
    if not verify:
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
        logger.warn("SSL certificate verification is disabled for Arkindex API calls")

    return ArkindexClient(verify=verify, **options)


# Time to wait before retrying the IIIF image information fetching
HTTP_GET_RETRY_BACKOFF = 10

DOWNLOAD_CHUNK_SIZE = 8192


@retry(
    reraise=True,
    retry=retry_if_exception_type(requests.RequestException),
    stop=stop_after_attempt(3),
    wait=wait_fixed(HTTP_GET_RETRY_BACKOFF),
)
def download_file(url: str, path: Path) -> None:
    """
    Download a URL into a local path, retrying if necessary
    """
    with requests.get(url, stream=True, verify=should_verify_cert(url)) as r:
        r.raise_for_status()
        with path.open("wb") as f:
            for chunk in r.iter_content(chunk_size=DOWNLOAD_CHUNK_SIZE):
                if chunk:  # Ignore empty chunks
                    f.write(chunk)
